@@ -59,12 +59,11 @@ use derivative::Derivative;
 mod client;
 mod config;
 mod error;
-mod events;
+pub mod events;
 pub mod extrinsic;
 mod metadata;
 pub mod rpc;
 pub mod storage;
-mod subscription;
 mod transaction;
 
 pub use crate::{
@@ -82,12 +81,14 @@ pub use crate::{
         BasicError,
         Error,
         GenericError,
+        HasModuleError,
         RuntimeError,
         TransactionError,
     },
     events::{
-        EventsDecoder,
-        RawEvent,
+        EventDetails,
+        Events,
+        RawEventDetails,
     },
     extrinsic::{
         DefaultExtra,
@@ -98,6 +99,7 @@ pub use crate::{
         UncheckedExtrinsic,
     },
     metadata::{
+        ErrorMetadata,
         Metadata,
         MetadataError,
         PalletMetadata,
@@ -114,11 +116,6 @@ pub use crate::{
         StorageEntryKey,
         StorageMapKey,
     },
-    subscription::{
-        EventStorageSubscription,
-        EventSubscription,
-        FinalizedEventStorageSubscription,
-    },
     transaction::{
         TransactionEvents,
         TransactionInBlock,
@@ -185,7 +182,8 @@ pub enum Phase {
     Clone(bound = ""),
     PartialEq(bound = ""),
     Eq(bound = ""),
-    Default(bound = "")
+    Default(bound = ""),
+    Hash(bound = "")
 )]
 pub struct WrapperKeepOpaque<T> {
     data: Vec<u8>,
@@ -197,7 +195,7 @@ impl<T: Decode> WrapperKeepOpaque<T> {
     ///
     /// Returns `None` if the decoding failed.
     pub fn try_decode(&self) -> Option<T> {
-        T::decode_all(&self.data[..]).ok()
+        T::decode_all(&mut &self.data[..]).ok()
     }
 
     /// Returns the length of the encoded `T`.
@@ -228,7 +226,8 @@ impl<T: Decode> WrapperKeepOpaque<T> {
     PartialEq(bound = ""),
     Debug(bound = ""),
     Eq(bound = ""),
-    Default(bound = "")
+    Default(bound = ""),
+    Hash(bound = "")
 )]
 #[scale_info(skip_type_params(T))]
 #[doc(hidden)]
@@ -242,3 +241,8 @@ impl<T> PhantomDataSendSync<T> {
 
 unsafe impl<T> Send for PhantomDataSendSync<T> {}
 unsafe impl<T> Sync for PhantomDataSendSync<T> {}
+
+/// This represents a key-value collection and is SCALE compatible
+/// with collections like BTreeMap. This has the same type params
+/// as `BTreeMap` which allows us to easily swap the two during codegen.
+pub type KeyedVec<K, V> = Vec<(K, V)>;
